html {
	margin: 0;
	padding: 0;
	background-color: #fff;
}

body {
	line-height: 1.4;
	max-width: 80ex;
	margin: 1em auto;
	background-color: #fff;
}

pre, code, samp, kbd {
	font-size: 0.9em;
}

kbd {
	padding: 2px;
	background-color: #f9f9f9;
	border-width: 1px;
	border-style: solid;
	font-style: revert;
	white-space: pre;
}

img {
	width: 100%;
}

sup {
	line-height: 0;
}

div.footnote h5 {
	display: table-cell;
	font-size: initial;
}

div.footnote p {
	display: table-cell;
	padding-left: 1em;
}

pre.menu-comment, pre.menu-preformatted {
	font-family: initial;
	margin: 0;
}


@media (prefers-color-scheme: dark) {
	html, body {
		background-color: #1e1f21;
		color: #eeeff1;
	}
	a { color: #bad7ff; }
	a:visited { color: #d099ff; }
	kbd { background-color: #404040 }
}
